import unittest

import fqr

from fqr . loggers import lib

from .. import mocking

from . import cfg


class Constants(cfg.Constants):
    """Constant values specific to this file."""


class TestLogger(unittest.TestCase):
    """Fixture for testing logger."""

    def setUp(self) -> None:
        self.log = fqr.log
        self.msg_str = 'example'
        self.msg_dict = {'str': 'example', 'a': 2}
        self.msg_cls = mocking.examples.Pet
        self.msg_object_= mocking.examples.Pet(
            id_='abc1234',
            name='Fido',
            type='dog',
            )
        return super().setUp()

    def test_01_log(self):
        """Test str logging."""

        msg = self.msg_str
        level = lib.logging.DEBUG
        expected_output = lib.textwrap.indent(
            lib.json.dumps(
                fqr.loggers.utl.parse_incoming_log_message(msg, level),
                default=fqr.core.strings.utl.convert_for_repr,
                indent=Constants.INDENT,
                sort_keys=True
                ),
            Constants.INDENT * ' '
            )
        with self.assertLogs(self.log, level) as logger:
            fqr.log.debug(msg)
            self.assertEqual(logger.records[0].msg, expected_output)

    def test_02_log(self):
        """Test dict logging."""

        msg = self.msg_dict
        level = lib.logging.INFO
        expected_output = lib.textwrap.indent(
            lib.json.dumps(
                fqr.loggers.utl.parse_incoming_log_message(msg, level),
                default=fqr.core.strings.utl.convert_for_repr,
                indent=Constants.INDENT,
                sort_keys=True
                ),
            Constants.INDENT * ' '
            )
        with self.assertLogs(self.log, level) as logger:
            fqr.log.info(msg)
            self.assertEqual(logger.records[0].msg, expected_output)

    def test_03_log(self):
        """Test cls logging."""

        msg = self.msg_cls
        level = lib.logging.WARNING
        expected_output = lib.textwrap.indent(
            lib.json.dumps(
                fqr.loggers.utl.parse_incoming_log_message(msg, level),
                default=fqr.core.strings.utl.convert_for_repr,
                indent=Constants.INDENT,
                sort_keys=True
                ),
            Constants.INDENT * ' '
            )
        with self.assertLogs(self.log, level) as logger:
            fqr.log.warning(msg)
            self.assertEqual(logger.records[0].msg, expected_output)

    def test_04_log(self):
        """Test object logging."""

        msg = self.msg_object_
        level = lib.logging.ERROR
        expected_output = lib.textwrap.indent(
            lib.json.dumps(
                fqr.loggers.utl.parse_incoming_log_message(msg, level),
                default=fqr.core.strings.utl.convert_for_repr,
                indent=Constants.INDENT,
                sort_keys=True
                ),
            Constants.INDENT * ' '
            )
        with self.assertLogs(self.log, level) as logger:
            fqr.log.error(msg)
            self.assertEqual(logger.records[0].msg, expected_output)

    def test_05_log(self):
        """Test exc logging."""

        msg = self.msg_str
        level = lib.logging.ERROR
        parsed = fqr.loggers.utl.parse_incoming_log_message(msg, level)
        with self.assertLogs(self.log, level) as logger:
            try:
                _ = 1 / 0
            except Exception:
                fqr.log.error(msg)
                expected_output = lib.textwrap.indent(
                    lib.json.dumps(
                        parsed,
                        default=fqr.core.strings.utl.convert_for_repr,
                        indent=Constants.INDENT,
                        sort_keys=True
                        ),
                    Constants.INDENT * ' '
                    )
            self.assertEqual(logger.records[0].msg, expected_output)

    def test_06_log(self):
        """Test exc logging."""

        msg = self.msg_str
        level = lib.logging.DEBUG
        parsed = fqr.loggers.utl.parse_incoming_log_message(msg, level)
        with self.assertLogs(self.log, level) as logger:
            try:
                _ = 1 / 0
            except Exception as e:
                self.log.debug(msg, exc_info=e)
                expected_output = lib.textwrap.indent(
                    lib.json.dumps(
                        parsed,
                        default=fqr.core.strings.utl.convert_for_repr,
                        indent=Constants.INDENT,
                        sort_keys=True
                        ),
                    Constants.INDENT * ' '
                    )
            self.assertEqual(logger.records[0].msg, expected_output)

    def test_07_print(self):
        """Test first print + warning interception."""

        msg = self.msg_str
        level = lib.logging.WARNING
        parsed = fqr.loggers.utl.parse_incoming_log_message(
            Constants.WARN_MSG,
            level
            )
        with self.assertLogs(self.log, level) as logger:
            print(msg)
            expected_output = lib.textwrap.indent(
                lib.json.dumps(
                    parsed,
                    default=fqr.core.strings.utl.convert_for_repr,
                    indent=Constants.INDENT,
                    sort_keys=True
                    ),
                Constants.INDENT * ' '
                )
            self.assertEqual(
                logger.records[0].msg,
                expected_output
                )

    def test_08_print(self):
        """Test subsequent print interceptions."""

        msg = self.msg_str
        with self.assertWarns(Warning) as emitter:
            with self.assertNoLogs(self.log):
                print(msg)
                self.assertEqual(
                    emitter.warnings[0].message.args[0],
                    Constants.WARN_MSG
                    )


class TestDeployedLogger(unittest.TestCase):
    """Fixture for testing logger in higher environments."""

    def setUp(self) -> None:
        Constants.ENV = 'dev'
        fqr.loggers.obj._set_print_interception()
        self.log = fqr.log
        self.msg = mocking.examples.Pet
        return super().setUp()

    def test_01_print(self):
        """Test first print + warning interception."""

        # lib.sys.__dict__['modules'][fqr.loggers.obj.__name__].__dict__.pop('__warningregistry__')

        msg = self.msg
        warn_msg = '\n'.join(
            (
                '.py:line_no: WarningClass: ',
                Constants.WARN_MSG,
                str(msg),
                'lib.warnings.warn()'
                )
            )
        level = lib.logging.WARNING
        parsed = fqr.loggers.utl.parse_incoming_log_message(warn_msg, level)
        with self.assertLogs(self.log, level) as logger:
            # expected_output = lib.textwrap.indent(
            #     lib.json.dumps(
            #         parsed,
            #         default=fqr.core.strings.utl.convert_for_repr,
            #         indent=Constants.INDENT,
            #         sort_keys=True
            #         ),
            #     Constants.INDENT * ' '
            #     )
            expected_output = lib.json.dumps(
                parsed,
                default=fqr.core.strings.utl.convert_for_repr,
                indent=Constants.INDENT,
                sort_keys=True
                )
            print(msg)
            self.assertEqual(
                logger.records[0].msg,
                expected_output
                )

    def test_02_print(self):
        """Test subsequent, same message print interceptions."""

        msg = self.msg
        warn_msg = '\n'.join(
            (
                Constants.SILENCE_MSG,
                str(msg),
                )
            )
        with self.assertWarns(Warning) as emitter:
            with self.assertNoLogs(self.log):
                print(msg)
                self.assertEqual(
                    emitter.warnings[0].message.args[0],
                    warn_msg
                    )

    def test_03_print(self):
        """Test subsequent, new message print + warning interception."""

        msg = 'testing'
        warn_msg = '\n'.join(
            (
                '.py:line_no: WarningClass: ',
                Constants.SILENCE_MSG,
                str(msg),
                'lib.warnings.warn()'
                )
            )
        level = lib.logging.WARNING
        parsed = fqr.loggers.utl.parse_incoming_log_message(warn_msg, level)
        with self.assertLogs(self.log, level) as logger:
            print(msg)
            expected_output = lib.textwrap.indent(
                lib.json.dumps(
                    parsed,
                    default=fqr.core.strings.utl.convert_for_repr,
                    indent=Constants.INDENT,
                    sort_keys=True
                    ),
                Constants.INDENT * ' '
                )
            self.assertEqual(
                logger.records[0].msg,
                expected_output
                )

    def test_04_ext_warning(self):
        """Test external warning interception."""

        msg = repr(self.msg)
        level = lib.logging.WARNING
        parsed = fqr.loggers.utl.parse_incoming_log_message(msg, level)
        with self.assertLogs(self.log, level) as logger:
            lib.warnings.warn(msg, stacklevel=1)
            expected_output = lib.textwrap.indent(
                lib.json.dumps(
                    parsed,
                    default=fqr.core.strings.utl.convert_for_repr,
                    indent=Constants.INDENT,
                    sort_keys=True
                    ),
                Constants.INDENT * ' '
                )
            self.assertEqual(
                logger.records[0].msg,
                expected_output
                )

    def test_05_invalid_log_type_exc(self):
        """Test correct exc is raised."""

        with self.assertNoLogs(self.log):
            self.assertRaises(
                fqr.loggers.exc.InvalidLogMessageTypeError,
                lambda: fqr.log.info(42)
                )

    def test_06_log_truncation(self):
        """Test long strings correctly truncated."""

        msg = {'longString': 'ABCDEFG' * 1024}
        level = lib.logging.CRITICAL
        with self.assertLogs(self.log, level) as logger:
            fqr.log.critical(msg)
            self.assertIn(
                fqr.core.strings.cfg.Constants.M_LINE_TOKEN,
                logger.records[0].msg
                )

    def test_07_log_redaction(self):
        """Test sensitive strings correctly redacted."""

        msg = {'apiKey': 'ABCDEFG'}
        level = lib.logging.INFO
        with self.assertLogs(self.log, level) as logger:
            fqr.log.info(msg)
            self.assertIn(
                '[ REDACTED :: API KEY ]',
                logger.records[0].msg
                )

    def tearDown(self) -> None:
        Constants.ENV = 'local'
        return super().tearDown()


class TestTracedLogger(unittest.TestCase):
    """Fixture for testing logger with tracing."""

    def setUp(self) -> None:
        Constants.LOG_TRACEBACK = True
        self.log = fqr.log
        self.msg_str = 'example'
        return super().setUp()

    def test_01_log(self):
        """Test exc logging."""

        msg = self.msg_str
        level = lib.logging.ERROR
        parsed = fqr.loggers.utl.parse_incoming_log_message(msg, level)
        with self.assertLogs(self.log, level) as logger:
            try:
                _ = 1 / 0
            except Exception:
                fqr.log.error(msg)
                parsed['traceback'] = lib.traceback.format_exc()
                expected_output = lib.textwrap.indent(
                    lib.json.dumps(
                        parsed,
                        default=fqr.core.strings.utl.convert_for_repr,
                        indent=Constants.INDENT,
                        sort_keys=True
                        ),
                    Constants.INDENT * ' '
                    )
            self.assertEqual(logger.records[0].msg, expected_output)

    def test_02_log(self):
        """Test exc logging - no trace for < ERROR."""

        msg = self.msg_str
        level = lib.logging.DEBUG
        parsed = fqr.loggers.utl.parse_incoming_log_message(msg, level)
        with self.assertLogs(self.log, level) as logger:
            try:
                _ = 1 / 0
            except Exception as e:
                self.log.info(msg, exc_info=e)
                expected_output = lib.textwrap.indent(
                    lib.json.dumps(
                        parsed,
                        default=fqr.core.strings.utl.convert_for_repr,
                        indent=Constants.INDENT,
                        sort_keys=True
                        ),
                    Constants.INDENT * ' '
                    )
            self.assertEqual(logger.records[0].msg, expected_output)

    def tearDown(self) -> None:
        Constants.LOG_TRACEBACK = False
        return super().tearDown()



"""Loggers constants."""

__all__ = (
    'Constants',
    )

from .. import core

from . import lib


class Constants(core.cfg.Constants):
    """Constant values shared across log modules."""

    FTIME_LOG      = '%Y-%m-%d %H:%M:%S'
    FTIME_LOG_MSEC = '%s.%03d UTC'

    LOG_LEVEL      = lib.os.getenv(
        'LOG_LEVEL',
        'DEBUG' if core.cfg.Constants.ENV in {'dev', 'develop', 'local'} else 'INFO'
        ).upper()
    LOG_TRACEBACK  = lib.os.getenv('LOG_TRACEBACK', 'false').lower() == 'true'
    LOG_PRINTS     = lib.os.getenv('LOG_PRINTS', 'false').lower() == 'true'

    SILENCE_MSG    = f'Call to print() silenced by {core.cfg.Constants.PACAKGE}.'
    WARN_MSG       = f'Calls to print() will be silenced by {core.cfg.Constants.PACAKGE}.'



"""Docs objects."""

__all__ = (
    'docs_parser',
    'Pattern',
    )

from .. import cli

from . import cfg
from . import enm
from . import lib
from . import utl

__all__ = (
    'docs_parser',
    'Pattern',
    )


class Constants(cfg.Constants):
    """Constant values specific to Docs objs."""


class Pattern:
    """Compiled regex patterns."""

    ModulePattern = lib.re.compile(r'^(\.\. automodule:: )')
    """Matches automodule:: directive text to be stripped from rst files."""

    MinOneWordPattern = lib.re.compile(r'\w+')
    """Matches at least one valid word."""

    PathPattern   = lib.re.compile(r'(\/|\\)')
    """Matches Windows and Unix paths."""

    PrefixPattern = lib.re.compile(r'^(\.)?(\/|\\)?((src)(\/|\\))?')
    """Matches packages that use the /src prefix."""


docs_parser = cli.obj.parsers.add_parser(  # type: ignore[has-type]
    'docs',
    formatter_class=lib.argparse.ArgumentDefaultsHelpFormatter,
    )

docs_parser.add_argument(
    'package',
    help='the name or path to the package to be documented',
    )
docs_parser.add_argument(
    '--version',
    '-v',
    default=Constants.BASE_VERSION,
    help='the version of the package',
    dest='version',
    )
docs_parser.add_argument(
    '--author',
    '-a',
    help='specify package author',
    dest='author',
    default=Constants.DEFAULT_USER,
    )
docs_parser.add_argument(
    '--output',
    '-o',
    help='specify directory in which to create docs/',
    dest='output_dir',
    default='.',
    )
docs_parser.add_argument(
    '--favicon',
    help='specify location of a favicon (.ico) file',
    dest='favicon_path',
    default=lib.os.path.join(Constants.DOCS_STATIC_DIR, 'favicon.ico'),
    )
docs_parser.add_argument(
    '--logo',
    help='specify location of a logo file',
    dest='logo_path',
    default=lib.os.path.join(Constants.DOCS_STATIC_DIR, 'logo.png'),
    )
docs_parser.add_argument(
    '--theme',
    choices=enm.SupportedTheme._member_names_,
    help='specify sphinx theme',
    dest='sphinx_theme',
    default=enm.SupportedTheme.alabaster.value,
    )
docs_parser.add_argument(
    '--namespace-package',
    action='store_true',
    dest='is_namespace_package',
    help='include to specify that package is a namespace package',
    )
docs_parser.add_argument(
    '--add-module-names',
    action='store_true',
    dest='add_module_names',
    help='include to prepend module names to object names',
    )
docs_parser.add_argument(
    '--no-inherit-docstrings',
    action='store_false',
    dest='autodoc_inherit_docstrings',
    help='include to disable docstring inheritance',
    )
docs_parser.add_argument(
    '--no-cleanup',
    action='store_true',
    dest='no_cleanup',
    help='include to disable /docs/source/ dir removal',
    )
docs_parser.add_argument(
    '--include-private-modules',
    action='store_true',
    dest='include_private_modules',
    help='include to also document _private modules',
    )
docs_parser.add_argument(
    '--make-index',
    action='store_true',
    dest='make_index',
    help='include to make root file called index instead of {package}',
    )
docs_parser.add_argument(
    '--index-from-readme',
    nargs='?',
    dest='readme_path',
    help=(
        'specify path to README.md file to use'
        ' for index instead of {package}'
        ),
    const=lib.os.path.join('.', 'README.md'),
    default=None,
    )
docs_parser.add_argument(
    '--no-include-meta-tags',
    action='store_true',
    dest='no_include_meta_tags',
    help='include to disable auto-generation of meta tags for documentation',
    )
docs_parser.add_argument(
    '--no-robots-txt',
    action='store_true',
    dest='no_include_robots',
    help='include to disable auto-generation of robots.txt file',
    )
docs_parser.add_argument(
    '--site-map-url',
    help='\n'.join(
        (
            (
                'specify full url path to the documentation version to be'
                'included in an auto-generated xml sitemap'
                ),
            '',
            'ex. https://example.readthedocs.io/en/stable',
            '',
            'can be repeated to include multiple versions'
            )
        ),
    action='append',
    dest='site_map_urls',
    )
docs_parser.add_argument(
    '--site-map-change-freq',
    choices=enm.SiteMapChangeFreq._member_names_,
    help='specify how often site-map expected to change',
    dest='site_map_change_freq',
    default=enm.SiteMapChangeFreq.weekly.value,
    )
docs_parser.set_defaults(func=utl.document)



"""Constant values specific to module unit tests."""

__all__ = (
    'Constants',
    )

from fqr import loggers


class Constants(loggers.cfg.Constants):
    """Constant values specific to unit tests in this module."""



"""Module utils unit tests."""

import unittest

import fqr

from ... import mocking

from . import cfg


class SimpleTypedObj(fqr.core.lib.t.TypedDict):
    """Simple `SupportsAnnotation` object."""

    name: str
    id_: int


class InitVarClass(fqr.Object):
    """Has an InitVar for testing."""

    init_var: fqr.Field[fqr.core.lib.dataclasses.InitVar]
    init_var_as_str: 'fqr.Field[fqr.core.lib.dataclasses.InitVar]'


class Constants(cfg.Constants):
    """Constant values specific to unit tests in this file."""


class TestUtils(unittest.TestCase):
    """Fixture for testing."""

    def test_01_str_to_forwardref(self):
        """Test `str` to `ForwardRef` casting."""

        self.assertIsInstance(
            fqr.core.typ.utl.hint.parse_str_to_ref('int', False),
            fqr.core.lib.t.ForwardRef
            )

    def test_02_str_to_type(self):
        """Test `str` to `type` casting."""

        self.assertIs(
            fqr.core.typ.utl.hint.resolve_type(
                'int',
                globals(),
                locals()
                ),
            int
            )

    def test_03_arg_ref_handling_no_ns(self):
        """Test `ForwardRef` handling for types not yet resolvable."""

        self.assertIsInstance(
            fqr.core.typ.utl.hint.resolve_type('Unresolvable'),
            fqr.core.lib.t.ForwardRef
            )

    def test_04_arg_ref_handling(self):
        """Test `ForwardRef` handling for types with args."""

        self.assertIs(
            fqr.core.typ.utl.hint.resolve_type(
                'Mockery[tuple[int, ...]]',
                globals(),
                locals()
                ),
            Mockery[tuple[int, ...]]
            )

    def test_05_anti_is_array_type(self):
        """Test `is_array_type`."""

        self.assertFalse(fqr.core.typ.utl.check.is_array_type(None))

    def test_06_anti_is_variadic_array_type(self):
        """Test `is_variadic_array_type`."""

        self.assertFalse(
            fqr.core.typ.utl.check.is_variadic_array_type(None)
            )

    def test_07_anti_is_mapping_type(self):
        """Test `is_mapping_type`."""

        self.assertFalse(fqr.core.typ.utl.check.is_mapping_type(None))

    def test_08_anti_is_none_type(self):
        """Test `is_none_type`."""

        self.assertFalse(fqr.core.typ.utl.check.is_none_type(None))

    def test_09_anti_is_number_type(self):
        """Test `is_number_type`."""

        self.assertFalse(fqr.core.typ.utl.check.is_number_type(None))

    def test_10_anti_is_bool_type(self):
        """Test `is_bool_type`."""

        self.assertFalse(fqr.core.typ.utl.check.is_bool_type(None))

    def test_11_anti_is_datetime_type(self):
        """Test `is_datetime_type`."""

        self.assertFalse(fqr.core.typ.utl.check.is_datetime_type(None))

    def test_12_anti_is_date_type(self):
        """Test `is_date_type`."""

        self.assertFalse(fqr.core.typ.utl.check.is_date_type(None))

    def test_13_annotations_cache(self):
        """Test `collect_annotations`."""

        self.assertEqual(
            SimpleTypedObj.__annotations__,
            fqr.core.typ.utl.hint.collect_annotations(SimpleTypedObj)
            )

    def test_14_is_object(self):
        """Test `is_object`."""

        self.assertTrue(
            fqr.core.typ.utl.check.is_object(mocking.Derivative)
            )

    def test_15_is_field_type_forward_str(self):
        """Test `is_field_type` works with `str` ForwardRef."""

        self.assertTrue(
            fqr.core.typ.utl.check.is_field_type('Field[Any]')
            )

    def test_16_is_immutable_type(self):
        """Test `is_immutable_type`."""

        self.assertTrue(
            fqr.core.typ.utl.check.is_immutable_type(
                tuple[str, ...]
                )
            )


class Mockery(fqr.core.lib.t.Generic[fqr.core.typ.AnyType]):
    """An as yet undefined generic class for testing."""



"""Core imports."""

__all__ = (
    'argparse',
    'dataclasses',
    'datetime',
    'decimal',
    'enum',
    'functools',
    'itertools',
    'json',
    'os',
    're',
    'sys',
    't',
    'textwrap',
    'types',
    'urllib',
    'LiteralString',
    'Never',
    'Self',
    'TypeVarTuple',
    'Unpack',
    )

import argparse
import dataclasses
import datetime
import decimal
import enum
import functools
import itertools
import json
import os
import re
import sys
import typing as t
import textwrap
import types
import urllib.parse

if sys.version_info < (3, 11):  # pragma: no cover
    from typing_extensions import (  # noqa  # type: ignore
        LiteralString,
        Never,
        Self,
        TypeVarTuple,
        Unpack
        )
else:  # pragma: no cover
    from typing import (
        LiteralString,
        Never,
        Self,
        TypeVarTuple,
        Unpack
        )



"""
Overview
========

**Summary:** fqr extension for logging.

---

Usage
-----

##### Controlled by the following environment variables / defaults.

```py
ENV = os.getenv('ENV', 'local').lower()
LOG_LEVEL = (
    os.getenv(
        'LOG_LEVEL',
        (
            'DEBUG'
            if ENV in {'dev', 'develop', 'local'}
            else 'INFO'
            )
        )
    ).upper()
# The default level for the logger.

LOG_PRINTS = os.getenv('LOG_PRINTS', 'false').lower() == 'true'
# Whether or not print() statements may be logged.

LOG_TRACEBACK = os.getenv('LOG_TRACEBACK', 'false').lower() == 'true'
# Whether or not error tracebacks may be logged.

```

"""

__all__ = (
    'cfg',
    'exc',
    'lib',
    'log',
    'obj',
    'typ',
    'utl',
    )

from . import cfg
from . import exc
from . import lib
from . import obj
from . import typ
from . import utl

from . obj import log



import pickle
import unittest

import fqr


class TestExceptions(unittest.TestCase):
    """Fixture for testing exceptions."""

    def setUp(self) -> None:
        return super().setUp()

    def test_01_serialization(self):
        """Test multi-arg exc serializes correctly."""

        exc = fqr.objects.exc.IncorrectCasingError(('a', 'b'))
        dump = pickle.dumps(exc)
        reloaded: fqr.core.typ.PackageExceptionType = pickle.loads(dump)
        self.assertTupleEqual(
            exc.args,
            reloaded.args
            )

    def test_02_serialization(self):
        """Test multi-arg exc serializes correctly."""

        exc = fqr.objects.exc.IncorrectDefaultTypeError('test', str, 2)
        dump = pickle.dumps(exc)
        reloaded: fqr.core.typ.PackageExceptionType = pickle.loads(dump)
        self.assertTupleEqual(
            exc.args,
            reloaded.args
            )

    def test_03_serialization(self):
        """Test multi-arg exc serializes correctly."""

        exc = fqr.objects.exc.IncorrectTypeError('test', str, 2)
        dump = pickle.dumps(exc)
        reloaded: fqr.core.typ.PackageExceptionType = pickle.loads(dump)
        self.assertTupleEqual(
            exc.args,
            reloaded.args
            )

    def test_04_serialization(self):
        """Test multi-arg exc serializes correctly."""

        exc = fqr.objects.exc.InvalidComparisonTypeError('test', str, 2)
        dump = pickle.dumps(exc)
        reloaded: fqr.core.typ.PackageExceptionType = pickle.loads(dump)
        self.assertTupleEqual(
            exc.args,
            reloaded.args
            )

    def test_05_serialization(self):
        """Test multi-arg exc serializes correctly."""

        exc = fqr.objects.exc.InvalidContainerComparisonTypeError(
            'test',
            list[str],
            2
            )
        dump = pickle.dumps(exc)
        reloaded: fqr.core.typ.PackageExceptionType = pickle.loads(dump)
        self.assertTupleEqual(
            exc.args,
            reloaded.args
            )

    def test_06_serialization(self):
        """Test exc serializes correctly."""

        exc = fqr.objects.exc.FieldAnnotationError('test', int)
        dump = pickle.dumps(exc)
        reloaded: fqr.core.typ.PackageExceptionType = pickle.loads(dump)
        self.assertTupleEqual(exc.args, reloaded.args)



"""Strings exceptions."""

from .. import exc

__all__ = (
    'StringCasingError',
    *exc.__all__
    )

from .. exc import *

from . import typ


class StringCasingError(BasePackageException[str, typ.Casing]):
    """Exception raised on invalid string casing."""

    def __init__(self, string: str, valid_case: typ.Casing) -> None:
        super().__init__(
            ' '.join(
                (
                    string,
                    'is not a valid',
                    f'`{valid_case!s}` string.'
                    )
                ),
            string,
            valid_case
            )
